use strict;
use vars qw($VERSION %IRSSI);

use Irssi;
use Data::Dumper;

$VERSION = '0.2';
%IRSSI = (
    authors     => 'cxreg',
    contact     => 'cxreg@pobox.com',
    name        => 'go',
    description => 'Switch to the window with the given name or item',
    license     => 'Public Domain',
    url         => 'http://genericorp.net/~count/irssi',
    changed     => '2008-01-29',
);

sub cmd_go {

    _draw_suggestion('');

    # get a complete list of current windows
    my @all_windows = Irssi::windows();

    my ($window, $suggestion, @matches);
    my $buf = '';
    while(1) {
        # read input one character at a time
        my $chr = getc;

        # break out on Enter
        if ( $chr =~ /[\r\n]/ ) {
            $window = $suggestion;
            last;
        }

        if ( $chr =~ /[\t]/ ) {
            # get matches if we don't have any yet
            @matches = _match( $buf, @all_windows ) unless @matches;

            # if tab, cycle to next suggestion
            push @matches, grep { $_ } shift @matches;
            $suggestion = $matches[0];

            # draw suggestion
            _draw_suggestion( $buf, $suggestion );

            next;
        }

        # handle backspace and delete
        if ( ord($chr) == 127 or ord($chr) == 8 ) {
            # remove the last char
            $buf = substr( $buf, 0, length($buf) - 1 );

            # get suggestions again
            if ( @matches = _match( $buf, @all_windows ) ) {
                $suggestion = $buf ? $matches[0] : undef;
                _draw_suggestion( $buf, $suggestion );
            }
            else {
                # we should only get here if the entire input was deleted
                $suggestion = undef;
            }

            next;
        }

        # create a temporary new buffer
        my $tmp = $buf . $chr;

        if ( @matches = _match($tmp, @all_windows) ) {
            $buf = $tmp;
            $suggestion = $buf ? $matches[0] : undef;
        } else {
            # vbell on mistype
            print STDOUT "\a";
        }

        # display the current input and suggestion
        _draw_suggestion( $buf, $suggestion );
    }

    if ($window) {
        $window->{window}->set_active;
    }

    # refresh the screen to get the regular prompt back if needed
    Irssi::command('redraw')
}

Irssi::command_bind('go', 'cmd_go', 'go.pl');

# load the command as a keybinding?
# Irssi::command('/bind meta-w /go');

sub _draw_suggestion {
    my ( $b, $s ) = @_;

    my ($pre, $post, $extra) = (''x3);
    if ($s) {
        if ( $s->{string} =~ /\Q$b\E/ ) {
            $pre = $`;
            $post = $';
        } elsif ( !$b ) {
            $post = $s->{string};
        }

        $extra = $s->{window}->{active_server}->{tag};
        $extra = " ($extra)" if $extra;
    }

    # FIXME - there has to be a "right way" to do this.  
    # it looks like the fe-text/gui-readline.c and gui-entry.c
    # (and other gui-*) are not XS wrapped for whatever reason.
    print STDOUT "\r" . ' 'x40 . "\rGoto: ";

    print STDOUT "\x{1b}[7m$pre\x{1b}[0m" if $pre; # before 
    print STDOUT $b;                               # the matched string
    print STDOUT "\x{1b}[7m$post$extra\x{1b}[0m"   # after
        if $post or $extra;
}

sub _match {
    my ( $name, @wins ) = @_;
    my @matches;

    for (@wins) {
        if ( length $_->{name} and ( $_->{name} =~ /\Q$name\E/ or !length($name) ) ) {
            push @matches, { string => $_->{name}, window => $_ };
        }
        if ( length $_->{active}->{name} and 
                ( $_->{active}->{name} =~ /\Q$name\E/ or !length($name) ) ) {
            push @matches, { string => $_->{active}->{name}, window => $_ };
        }
    }
    return @matches;
}
